
#include "../../DirectSoundPlayer.h"
#include "../../DirectSoundDriver.h"

#include "../ExtensibleFormat.h"

#include "DirectSoundPrivate.h"

#include <math.h>


namespace Juna {
namespace Sound {
using namespace DirectSound;


bool DirectSoundPoolPlayer::Start(void)
{
	return SUCCEEDED(lpDSBuffer->Play(0,0,LoopFlag ? DSBPLAY_LOOPING : 0));
}

bool DirectSoundPoolPlayer::Stop(void)
{
	return SUCCEEDED(lpDSBuffer->Stop()) &&
		   SUCCEEDED(lpDSBuffer->SetCurrentPosition(0));
}

bool DirectSoundPoolPlayer::isPlaying(void) const
{
	DWORD status = 0;
	return SUCCEEDED(lpDSBuffer->GetStatus(&status)) && (status & DSBSTATUS_PLAYING);
}

const I_PlayDriver *DirectSoundPoolPlayer::GetDriver(void)
{return &DirectSoundPlayDriver::GetInstance();}


void DirectSoundPoolPlayer::SetLoopFlag(bool loop)
{
	if (isPlaying())
		lpDSBuffer->Play(0,0,loop ? DSBPLAY_LOOPING : 0);
	LoopFlag = loop;
}


long DirectSoundPoolPlayer::GetPlayPosition(void)
{
	DWORD play;
//obt@̍Đʒu߂
	if FAILED(lpDSBuffer->GetCurrentPosition(&play,NULL))
		return -1;
//ByteSampleɕϊ邽߃tH[}bg߂
	return play / PlayFormat.sample_size();
}

bool DirectSoundPoolPlayer::SetPlayPosition(long iSamplePos)
{
	return SUCCEEDED(lpDSBuffer->SetCurrentPosition(iSamplePos * PlayFormat.sample_size()));
}



//_ŉʐݒ[1.0 - 0.00001]
bool DirectSoundPoolPlayer::SetVolumef(float left,float right)
{
//0.0log100Ԃ₪邩玩ŒׂĂ݂B
//Ȃ݂SetVolumeoRȂ΂Ƀ}CiXɂ͂ȂȂ璲ׂȂB
	if (max(left,right) <= 0.00001)
		return SUCCEEDED(lpDSBuffer->SetVolume(-10000));

//ΐnfVxɕϊAłĂ̂ȁH
	LONG dleft = static_cast<LONG>(100.0 * 20.0 * log10(left));
	LONG dright = static_cast<LONG>(100.0 * 20.0 * log10(right));
	if FAILED(lpDSBuffer->SetVolume(min(max(dleft,dright),0)))
		return false;
	if (PlayFormat.channel > 1 && FAILED(lpDSBuffer->SetPan(dright - dleft)))
	{}
	return true;
}
bool DirectSoundPoolPlayer::GetVolumef(float &left,float &right) const
{
	LONG dleft,dright;
	if FAILED(lpDSBuffer->GetVolume(&dleft))
		return false;
	LONG pan;
	if (PlayFormat.channel == 1 || FAILED(lpDSBuffer->GetPan(&pan)))
	{
//fVxwaveOutnŎgĂΐnɕϊ	ŗǂ̂H悭
		left = right =
			static_cast<float>(pow(10.0,(static_cast<double>(dleft) / 100.0) / 20.0));
		return true;
	}
	dright = dleft;
	if (pan > 0)
		dleft -= pan;
	else
		dright += pan;
	left = static_cast<float>(pow(10.0,(static_cast<double>(dleft) / 100.0) / 20.0));
	right = static_cast<float>(pow(10.0,(static_cast<double>(dright) / 100.0) / 20.0));
	return true;
}


bool DirectSoundPoolPlayer::SetVolume_dB(long left,long right)
{
	if FAILED(lpDSBuffer->SetVolume(max(max(left,right),DSBVOLUME_MIN)))
		return false;
	if (PlayFormat.channel > 1 && FAILED(lpDSBuffer->SetPan(right - left)))
	{}
	return true;
}
bool DirectSoundPoolPlayer::GetVolume_dB(long &left,long &right) const
{
	if FAILED(lpDSBuffer->GetVolume(&left))
		return false;
	right = left;
	LONG pan;
	if (PlayFormat.channel == 1 || FAILED(lpDSBuffer->GetPan(&pan)))
		return true;

	if (pan > 0)
		left -= pan;
	else
		right += pan;
	return true;
}

unsigned long DirectSoundPoolPlayer::GetBufferSize(void) const
{
	DSBCAPS dsbc = {sizeof(dsbc)};
	lpDSBuffer->GetCaps(&dsbc);

	return dsbc.dwBufferBytes / PlayFormat.sample_size();
}


I_DirectSoundPoolPlayer *DirectSoundPoolPlayer::Initialize
	(int device,const Format &format,long bufsize,int focus)
{
	if (lpDSBuffer)
	{
		lpDSBuffer->Release();
		lpDSBuffer = NULL;
	}

	if (Master) {
		Master->Release();
		Master = NULL;
	}
	{	//DirectSoundObject̊l
		const GUID *guid = DirectSoundPlayDriver::GetInstance().GetDSDeviceGUID(device);
		if (guid == NULL)
			goto ERR;

		DirectSoundObject *dso = DirectSoundObjectManager::GetInstance().GetDSObject(*guid);
		if (dso == NULL)
		{
			dso = DirectSoundObjectManager::GetInstance().InitializeDSObject(*guid,format);
			if (dso == NULL)
				goto ERR;
		}
		Master = dso;
	}

	if (!CreateBuffer(format,bufsize,focus))
		goto ERR;

	PlayFormat = format;
	return this;

ERR:;
	Finalize();
	return 0;
}


bool DirectSoundPoolPlayer::CreateBuffer
	(const Format &format,long bufsize,int focus)
{
	if (lpDSBuffer != NULL)
		return false;

	if (bufsize == 0)
		return false;
	if ((format.bit == 8 && format.quantization == QT_SIGNED) ||
		(format.bit != 8 && format.quantization == QT_UNSIGNED))
		return false;

//TvTCYɍ킹
	if (bufsize % format.sample_size())
		bufsize -= bufsize % format.sample_size();

	DWORD dsbdflag = DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLPAN;
//tH[JXݒ
	if (focus < 0)
		dsbdflag |= DSBCAPS_STICKYFOCUS;
	else if (focus > 0)
		dsbdflag |= DSBCAPS_GLOBALFOCUS;

	WAVEFORMATEX wfx;
	SetWaveFormatEx(wfx,format);

	DSBUFFERDESC dsbd = {sizeof(dsbd),dsbdflag,bufsize,0,&wfx};

	LPDIRECTSOUNDBUFFER lpBuffer = NULL;
	if (format.bit > 16 || format.channel > 2 ||
		FAILED(Master->GetDSObject()->CreateSoundBuffer(&dsbd,&lpBuffer,NULL)))
	{	//ExtensibleFormat
		const int extsize = sizeof(WAVEFORMATEX) + 22;
		BYTE buf[extsize];
		WAVEFORMATEX *wfext = reinterpret_cast<WAVEFORMATEX *>(buf);
		::memcpy(wfext,&wfx,sizeof(WAVEFORMATEX));
		wfext->cbSize = extsize - sizeof(WAVEFORMATEX);
		if (!MakeExtensibleFormat(wfext))
			return false;
		dsbd.lpwfxFormat = wfext;
		if FAILED(Master->GetDSObject()->CreateSoundBuffer(&dsbd,&lpBuffer,NULL))
			return false;
	}
	lpDSBuffer = lpBuffer;

	return true;
}
I_DirectSoundPoolPlayer *DirectSoundPoolPlayer::Initialize
	(int device,I_SeekableInputStream &stream,int focus)
{
	Format format = stream.GetFormat();
	long bufsize = stream.GetLength() * format.sample_size();
	if (!Initialize(device,format,bufsize,focus))
		return 0;
	if (stream.Seek(0) != 0 || !LoadStream(stream))
	{
		Finalize();
		return 0;
	}
	return this;
}

I_DirectSoundPoolPlayer *DirectSoundPoolPlayer::Initialize(I_PoolPlayer &duplsrc)
{
	Finalize();
	if (GetDriver() != duplsrc.GetDriver())
		return 0;

	DirectSoundPoolPlayer &dsp = static_cast<DirectSoundPoolPlayer &>(duplsrc);

	LPDIRECTSOUND dsobj = dsp.Master->GetDSObject();

	if FAILED(dsobj->DuplicateSoundBuffer(dsp.lpDSBuffer,&lpDSBuffer))
	{
		lpDSBuffer = NULL;
		return 0;
	}
	Master = dsp.Master;
	Master->AddRef();

	PlayFormat = dsp.PlayFormat;
	LoopFlag = dsp.LoopFlag;

	return this;
}


void DirectSoundPoolPlayer::Finalize(void)
{
	if (lpDSBuffer) {
		lpDSBuffer->Release();
		lpDSBuffer = NULL;
	}
	if (Master) {
		Master->Release();
		Master = NULL;
	}
	PlayFormat.reset();
}

bool DirectSoundPoolPlayer::LoadStream(I_InputStream &stream)
{
	if (!Stop())
		return false;
	if (PlayFormat != stream.GetFormat())
		return false;

	void *lock;
	unsigned long bufsize = GetBufferSize();
	if FAILED(lpDSBuffer->Lock(0,bufsize,&lock,&bufsize,NULL,NULL,DSBLOCK_ENTIREBUFFER))
		return false;

	unsigned long loadsize = stream.Read(lock,bufsize);
	int fill = (PlayFormat.bit == 8) ? 128 : 0;
	::memset(static_cast<char *>(lock) + loadsize,fill,bufsize - loadsize);

	lpDSBuffer->Unlock(lock,bufsize,NULL,0);

	return true;
}

}//namespace Sound
}//namespace Juna

