
#include "../../waveOutDriver.h"
#include "../../waveOutPlayer.h"

#include "../ExtensibleFormat.h"

namespace Juna {
namespace Sound
{

waveOutDeviceData *waveOutPlayDriver::list = NULL;
waveOutDeviceDataW *waveOutPlayDriver::wlist = NULL;

waveOutPlayDriver::waveOutPlayDriver(void) {}

waveOutPlayDriver::~waveOutPlayDriver() {DeleteDeviceList();}


void waveOutPlayDriver::DestroyPlayer(I_StreamPlayer *player)
{
	if (!player || player->GetDriver() != this)
		return;
	delete static_cast<base_waveOutStreamPlayer *>(player);
}
void waveOutPlayDriver::DestroyPlayer(I_PoolPlayer *player)
{
	if (!player || player->GetDriver() != this)
		return;
	delete static_cast<base_waveOutPoolPlayer *>(player);
}


I_waveOutStreamPlayer *waveOutPlayDriver::CreateWOStreamPlayer
	(int iDevice,const Format &iFormat,long iBufsize,int iBufnum)
{
	waveOutStreamPlayer *p = new waveOutStreamPlayer();
	if (p && !p->Initialize(iDevice,iFormat,iBufsize,iBufnum))
	{
		delete p;
		return 0;
	}
	return p;
}
I_waveOutStreamPlayer *waveOutPlayDriver::CreateWOStreamPlayer
	(int iDevice,I_InputStream &setStream,long iBufsize,int iBufnum)
{
	waveOutStreamPlayer *p = new waveOutStreamPlayer();
	if (p && !p->Initialize(iDevice,setStream,iBufsize,iBufnum))
	{
		delete p;
		return 0;
	}
	return p;
}
I_waveOutPoolPlayer *waveOutPlayDriver::CreateWOPoolPlayer
	(int iDevice,const Format &iFormat,long iBufSize)
{
	waveOutPoolPlayer *p = new waveOutPoolPlayer();
	if (p && !p->Initialize(iDevice,iFormat,iBufSize))
	{
		delete p;
		return 0;
	}
	return p;
}
I_waveOutPoolPlayer *waveOutPlayDriver::CreateWOPoolPlayer
	(int iDevice,I_SeekableInputStream &iStream)
{
	waveOutPoolPlayer *p = new waveOutPoolPlayer();
	if (p && !p->Initialize(iDevice,iStream))
	{
		delete p;
		return 0;
	}
	return p;
}
I_waveOutPoolPlayer *waveOutPlayDriver::CreateWOPoolPlayer
	(I_PoolPlayer &iDuplicateSrc)
{
	waveOutPoolPlayer *p = new waveOutPoolPlayer();
	if (p && !p->Initialize(iDuplicateSrc))
	{
		delete p;
		return 0;
	}
	return p;
}


I_waveOutStreamPlayer *waveOutExVolumePlayDriver::CreateWOStreamPlayer
	(int iDevice,const Format &iFormat,long iBufsize,int iBufnum)
{
	waveOutExVolumeStreamPlayer *p = new waveOutExVolumeStreamPlayer();
	if (p && !p->Initialize(iDevice,iFormat,iBufsize,iBufnum))
	{
		delete p;
		return 0;
	}
	return p;
}
I_waveOutStreamPlayer *waveOutExVolumePlayDriver::CreateWOStreamPlayer
	(int iDevice,I_InputStream &setStream,long iBufsize,int iBufnum)
{
	waveOutExVolumeStreamPlayer *p = new waveOutExVolumeStreamPlayer();
	if (p && !p->Initialize(iDevice,setStream,iBufsize,iBufnum))
	{
		delete p;
		return 0;
	}
	return p;
}
I_waveOutPoolPlayer *waveOutExVolumePlayDriver::CreateWOPoolPlayer
	(int iDevice,const Format &iFormat,long iBufSize)
{
	waveOutExVolumePoolPlayer *p = new waveOutExVolumePoolPlayer();
	if (p && !p->Initialize(iDevice,iFormat,iBufSize))
	{
		delete p;
		return 0;
	}
	return p;
}
I_waveOutPoolPlayer *waveOutExVolumePlayDriver::CreateWOPoolPlayer
	(int iDevice,I_SeekableInputStream &iStream)
{
	waveOutExVolumePoolPlayer *p = new waveOutExVolumePoolPlayer();
	if (p && !p->Initialize(iDevice,iStream))
	{
		delete p;
		return 0;
	}
	return p;
}
I_waveOutPoolPlayer *waveOutExVolumePlayDriver::CreateWOPoolPlayer
	(I_PoolPlayer &iDuplicateSrc)
{
	waveOutExVolumePoolPlayer *p = new waveOutExVolumePoolPlayer();
	if (p && !p->Initialize(iDuplicateSrc))
	{
		delete p;
		return 0;
	}
	return p;
}




void waveOutPlayDriver::UpdateDeviceList(void)
{
	DeleteDeviceList();
	int count = waveOutGetNumDevs();
	count = count + (count > 0);
	if (count == 0) return;
	list = new waveOutDeviceData[count];
	for (int i = 0;i < count;i++)
	{
		waveOutGetDevCapsA(i-1,&list[i].wocaps,sizeof(WAVEOUTCAPSA));
		list[i].ID = i - 1;
		list[i].name = list[i].wocaps.szPname;
	}
}
void waveOutPlayDriver::UpdateDeviceListW(void)
{
	DeleteDeviceList();
	int count = waveOutGetNumDevs();
	count = count + (count > 0);
	if (count == 0) return;
	wlist = new waveOutDeviceDataW[count];
	for (int i = 0;i < count;i++)
	{
		waveOutGetDevCapsW(i-1,&wlist[i].wocaps,sizeof(WAVEOUTCAPSW));
		wlist[i].ID = i - 1;
		wlist[i].name = wlist[i].wocaps.szPname;
	}
}void waveOutPlayDriver::DeleteDeviceList(void)
{
	if (list)
		delete[] list;
	if (wlist)
		delete[] wlist;
	list = NULL;
	wlist = NULL;
}

int waveOutPlayDriver::GetDeviceNum(void)
{
	int count = waveOutGetNumDevs();
	return count + (count > 0);
}

const PlayDeviceData *waveOutPlayDriver::GetDeviceData(int index)
{return GetWODeviceData(index);}

const waveOutDeviceData *waveOutPlayDriver::GetWODeviceData(int index)
{
	if (list == NULL) UpdateDeviceList();
	if (index < 0 || index >= GetDeviceNum())
		return NULL;
	return &list[index];
}

const PlayDeviceDataW *waveOutPlayDriver::GetDeviceDataW(int index)
	{return GetWODeviceDataW(index);}

const waveOutDeviceDataW *waveOutPlayDriver::GetWODeviceDataW(int index)
{
	if (wlist == NULL) UpdateDeviceListW();
	if (index < 0 || index >= GetDeviceNum())
		return NULL;
	return &wlist[index];
}

const UINT *waveOutPlayDriver::GetWODeviceID(int index)
{
	if (wlist)
	{
		const waveOutDeviceDataW *dd = GetInstance().GetWODeviceDataW(index);
		if (dd == NULL)
			return NULL;
		return &dd->ID;
	}
	else
	{
		const waveOutDeviceData *dd = GetInstance().GetWODeviceData(index);
		if (dd == NULL)
			return NULL;
		return &dd->ID;
	}
}

bool waveOutPlayDriver::TestOpenDevice(int index,const Format &format)
{
	if ((format.bit == 8 && format.quantization == QT_SIGNED) ||
		(format.bit != 8 && format.quantization == QT_UNSIGNED))
		return false;

	const UINT *id = GetWODeviceID(index);
	if (id == NULL)
		return false;

	WAVEFORMATEX wfx;
	SetWaveFormatEx(wfx,format);

	HWAVEOUT hwo;
	DWORD openflag = CALLBACK_NULL;
	if (format.bit > 16 || format.channel > 2 ||
		waveOutOpen(&hwo,*id,&wfx,0,0,openflag) != MMSYSERR_NOERROR)
	{
		const int extsize = sizeof(WAVEFORMATEX) + 22;
		BYTE buf[extsize];
		WAVEFORMATEX *wfext = reinterpret_cast<WAVEFORMATEX *>(buf);
		memcpy(wfext,&wfx,sizeof(WAVEFORMATEX));
		wfext->cbSize = extsize - sizeof(WAVEFORMATEX);
		if (!MakeExtensibleFormat(wfext))
			return false;
		if (waveOutOpen(&hwo,*id,wfext,0,0,openflag) != MMSYSERR_NOERROR)
			return false;
	}
	waveOutClose(hwo);
	return true;
}


waveOutPlayDriver &waveOutPlayDriver::GetInstance(void)
{
	static waveOutPlayDriver Driver;
	return Driver;
}



waveOutExVolumePlayDriver &waveOutExVolumePlayDriver::GetInstance(void)
{
	static waveOutExVolumePlayDriver Driver;
	return Driver;
}




I_waveOutPlayDriver *GetWaveOutPlayDriver(void)
{
	return &waveOutPlayDriver::GetInstance();
}

I_waveOutPlayDriver *GetWaveOutExVolumePlayDriver(void)
{
	return &waveOutExVolumePlayDriver::GetInstance();
}


}//namespace Sound
}//namespace Juna
