
#include "../../waveOutPlayer.h"
#include "../../waveOutDriver.h"

#include "../ExtensibleFormat.h"

#pragma comment(lib, "winmm.lib")

#include <process.h>

#include "EmptySupplier.h"

namespace Juna {
namespace Sound {


namespace {
class Synchronizer
{
public:
	Synchronizer(CRITICAL_SECTION &locker) : Locker(locker)
	{::EnterCriticalSection(&Locker);}
	~Synchronizer(void)
	{::LeaveCriticalSection(&Locker);}
private:
	CRITICAL_SECTION &Locker;
};

class NullCallbackClass : public I_StreamPlayer::I_PlayEndCallback
{
public: virtual void PlayEndCallback(void) {}
};

NullCallbackClass NullCallback; 

}//namespace ""


base_waveOutStreamPlayer::base_waveOutStreamPlayer(void)
  : hWaveOut(NULL),Stream(&PlayerEmptyStream),
	WaveHdr(NULL),BufferSize(0),BufferNum(0),WriteBufferPos(0),
	ThreadState(TS_EXIT),hDoneEvent(NULL),hThread(NULL),LoadEnd(FALSE),
	PlayFormat(),
	LeftVolume(0x10000),RightVolume(0x10000),VolumeFlag(false),
	PlayEndCallback(&NullCallback)
{
	::InitializeCriticalSection(&CritSect);
}
base_waveOutStreamPlayer::~base_waveOutStreamPlayer()
{
	::DeleteCriticalSection(&CritSect);
}

unsigned __stdcall base_waveOutStreamPlayer::PlayThread(void *woPlayer)
{
	base_waveOutStreamPlayer &this_ = *static_cast<base_waveOutStreamPlayer *>(woPlayer);

	while (1)
	{
		::WaitForSingleObject(this_.hDoneEvent,INFINITE);

		Synchronizer sync(this_.CritSect);
		switch (this_.ThreadState)
		{
		case TS_EXIT:
			return 0;
		case TS_PLAY:
			if (this_.LoadStream() == 0)
				this_.LoadStream();
			break;
		case TS_PAUSE:
			::waveOutPause(this_.hWaveOut);	//~
			if (this_.LoadEnd)
				this_.PlayEndCallback->PlayEndCallback();
			break;
		}
	}
//BR[h
	return 0;
}


void CALLBACK base_waveOutStreamPlayer::waveOutProc
	(HWAVEOUT hwo,UINT uMsg,DWORD dwInstance,DWORD dwParam1,DWORD dwParam2)
{
	base_waveOutStreamPlayer &this_ = *reinterpret_cast<base_waveOutStreamPlayer *>(dwInstance);
	switch (uMsg)
	{
	case WOM_OPEN:
		break;
	case WOM_CLOSE:
		break;
	case WOM_DONE:
//ʃXbhɃCxgŒʐMB
		if (this_.ThreadState == TS_PLAY)
		{
			if (this_.LoadEnd)	//[h
			{
				LPWAVEHDR lpwh;
				lpwh = (LPWAVEHDR)dwParam1;
				if (lpwh->dwBufferLength < (DWORD)this_.BufferSize)//ŏIf[^
				{
					this_.ThreadState = TS_PAUSE;
					SetEvent(this_.hDoneEvent);	//obt@̍ĐԂ~ɂ悤Ɉ˗
				}
			}
			else
				SetEvent(this_.hDoneEvent);		//f[^[h˗
		}
		break;
	}
}

const I_PlayDriver *base_waveOutStreamPlayer::GetDriver(void)
{return &waveOutPlayDriver::GetInstance();}


bool base_waveOutStreamPlayer::Start(void)
{
	Synchronizer sync(CritSect);
	if (ThreadState == TS_STANDBY && WriteBufferPos == 0)
		if (LoadStream() == -1)	//ŏ̈̃f[^[h
			return false;
	ThreadState = TS_PLAY;
	MMRESULT r = ::waveOutRestart(hWaveOut);
	::SetEvent(hDoneEvent);		//ʃXbhɎ̃f[^[h˗

	return r == MMSYSERR_NOERROR;
}

bool base_waveOutStreamPlayer::Stop(void)
{
	Synchronizer sync(CritSect);
	if (ThreadState == TS_PLAY)
	{
		::waveOutPause(hWaveOut);
		ThreadState = TS_PAUSE;
	}
	return true;
}

bool base_waveOutStreamPlayer::isPlaying(void) const
{
//	Synchronizer sync(CritSect);
	return (ThreadState & TS_PLAY) != 0;
}

long base_waveOutStreamPlayer::GetPlayTime(void)
{
//	Synchronizer sync(CritSect);
	MMRESULT mmr;
	MMTIME mmt;
	mmt.wType = TIME_SAMPLES;
	mmr = ::waveOutGetPosition(hWaveOut,&mmt,sizeof(mmt));
	if (mmr == MMSYSERR_NOERROR)
		return mmt.u.sample;
	return -1;
}

void base_waveOutStreamPlayer::SetPlayEndCallback(I_PlayEndCallback *setCallback)
{
	if (setCallback)
		PlayEndCallback = setCallback;
	else
		PlayEndCallback = &NullCallback;
}



bool base_waveOutStreamPlayer::SetDeviceVolume(unsigned long left,unsigned long right)
{
//	Synchronizer sync(CritSect);
	left = min(left,0xFFFF);
	right = min(right,0xFFFF);
	DWORD setvol = left | (right << 16);
	return ::waveOutSetVolume(hWaveOut,setvol) == MMSYSERR_NOERROR;
}

bool base_waveOutStreamPlayer::GetDeviceVolume(unsigned long &left,unsigned long &right) const
{
//	Synchronizer sync(CritSect);
	DWORD getvol;
	if (::waveOutGetVolume(hWaveOut,&getvol) == MMSYSERR_NOERROR)
	{
		right = (getvol >> 16) & 0xFFFF;
		left = getvol & 0xFFFF;
		right += (right == 0xFFFF);
		left += (left == 0xFFFF);
		return true;
	}
	return false;
}

bool base_waveOutStreamPlayer::SetSupplier(I_InputStream *setStream)
{
	if (!isInitialized())
		return false;

	Stop();
	Synchronizer sync(CritSect);

	ThreadState = TS_STANDBY;
	::waveOutReset(hWaveOut);
	::waveOutPause(hWaveOut);
	WriteBufferPos = 0;
	LoadEnd = 0;

	if (setStream == Stream)
		return true;
	if (setStream)
	{
		if (PlayFormat != setStream->GetFormat())
			return false;
	}
	else
		setStream = &PlayerEmptyStream;

	setStream->AddRef();
	Stream->Release();
	Stream = setStream;
	return true;
}


bool base_waveOutStreamPlayer::SetupDevice
	(int device,const Format &format,long bufsize,int bufnum)
{
	if ((format.bit == 8 && format.quantization == QT_SIGNED) ||
		(format.bit != 8 && format.quantization == QT_UNSIGNED))
		return false;

	const UINT *id = waveOutPlayDriver::GetInstance().GetWODeviceID(device);
	if (id == NULL)
		return false;

	WAVEFORMATEX wfx;
	SetWaveFormatEx(wfx,format);
	DWORD openflag = CALLBACK_FUNCTION;
//	if (dd->ID == WAVE_MAPPER) openflag |= WAVE_MAPPED;
	HWAVEOUT hwo;
	if (format.bit > 16 || format.channel > 2 ||
		::waveOutOpen(&hwo,*id,&wfx,(DWORD)waveOutProc,(DWORD)this,openflag) != MMSYSERR_NOERROR)
	{
		const int extsize = sizeof(WAVEFORMATEX) + 22;
		BYTE buf[extsize];
		WAVEFORMATEX *wfext = reinterpret_cast<WAVEFORMATEX *>(buf);
		memcpy(wfext,&wfx,sizeof(WAVEFORMATEX));
		wfext->cbSize = extsize - sizeof(WAVEFORMATEX);
		if (!MakeExtensibleFormat(wfext))
			return false;
		if (::waveOutOpen(&hwo,*id,wfext,(DWORD)waveOutProc,(DWORD)this,openflag) != MMSYSERR_NOERROR)
			return false;
	}
	hWaveOut = hwo;
	::waveOutPause(hWaveOut);	//ȂȂ悤ɒ~

#define Default_MSec (1000)
//ftHgł0w肳ƎIɌ
	if (bufsize == 0)
		bufsize = format.frequency * format.sample_size() * Default_MSec / 1000;
//TvTCYɍ킹
	if (bufsize % format.sample_size())
		bufsize -= bufsize % format.sample_size();
//Ȃ񂩃[mۂł̓G[łȂ`FbN
	if (bufnum == 0)
		return false;
	BufferSize = bufsize;
	BufferNum = bufnum;

//ŏɈCɃm
	WaveHdr = (WAVEHDR *)operator new((sizeof(WAVEHDR) + bufsize) * bufnum);
	if (WaveHdr == NULL)
		return false;

	for (int i = 0;i < bufnum;i++)
	{
		WaveHdr[i].lpData = ((LPSTR)(WaveHdr + bufnum)) + i * bufsize;
		WaveHdr[i].dwBufferLength = bufsize;
		WaveHdr[i].dwFlags = WHDR_DONE;
	}
	PlayFormat = format;

	return true;
}

bool base_waveOutStreamPlayer::base_Initialize
	(int device,const Format &iFormat,long bufsize,int bufnum)
{
	if (hWaveOut)
	{
		SetSupplier(0);
		::waveOutReset(hWaveOut);
		for (int i = 0;i < BufferNum;i++)
			::waveOutUnprepareHeader(hWaveOut,WaveHdr + i,sizeof(WAVEHDR));
		operator delete(WaveHdr);
		WaveHdr = NULL;
		::waveOutClose(hWaveOut);
		hWaveOut = NULL;
	}
	if (hDoneEvent == NULL)
	{
		hDoneEvent = ::CreateEvent(NULL,FALSE,FALSE,NULL);
		if (hDoneEvent == NULL)
			goto ERR;
	}

	if (!SetupDevice(device,iFormat,bufsize,bufnum))
		goto ERR;

	if (hThread == NULL)
	{
		ThreadState = TS_STANDBY;
		unsigned threadID;
		hThread = (HANDLE)::_beginthreadex(NULL,0,PlayThread,this,0,&threadID);
		if (hThread == NULL)
			goto ERR;
		::SetThreadPriority(hThread,THREAD_PRIORITY_ABOVE_NORMAL);
	}

	return true;

ERR:;
	base_Terminalize();
	return false;
}


void base_waveOutStreamPlayer::base_Terminalize(void)
{
	if (hWaveOut)
		SetSupplier(0);
	if (hThread)
	{
		do
		{
			ThreadState = TS_EXIT;		//Ƃ肠Đ~߂
			::SetEvent(hDoneEvent);	//蓮őҋ@Ԃ
//Î҂AIĂȂx
		} while (::WaitForSingleObject(hThread,10) == WAIT_TIMEOUT);
		::CloseHandle(hThread);
		hThread = NULL;
	}
	if (hDoneEvent)
	{
		::CloseHandle(hDoneEvent);
		hDoneEvent = NULL;
	}
	if (WaveHdr)
	{
		::waveOutReset(hWaveOut);
		for (int i = 0;i < BufferNum;i++)
			::waveOutUnprepareHeader(hWaveOut,WaveHdr + i,sizeof(WAVEHDR));
		operator delete(WaveHdr);
		WaveHdr = NULL;
	}
	if (hWaveOut)
	{
		::waveOutClose(hWaveOut);
		hWaveOut = NULL;
	}

	PlayFormat.reset();
}





/*=============================================================================

=============================================================================*/

bool waveOutStreamPlayer::Initialize
	(int device,const Format &format,long bufsize,int bufnum)
{return base_Initialize(device,format,bufsize,bufnum);}

bool waveOutStreamPlayer::Initialize
	(int device,I_InputStream &stream,long bufsize,int bufnum)
{
	if (!Initialize(device,stream.GetFormat(),bufsize,bufnum))
		return false;
	if (!SetSupplier(&stream))
	{
		Terminalize();
		return false;
	}
	return true;
}

void waveOutStreamPlayer::Terminalize(void) {base_Terminalize();}

int waveOutStreamPlayer::LoadStream(void)
{
	if (WaveHdr[WriteBufferPos].dwFlags & WHDR_DONE)
	{
		unsigned long size = Stream->Read(WaveHdr[WriteBufferPos].lpData,BufferSize);
		WaveHdr[WriteBufferPos].dwBufferLength = size;
		WaveHdr[WriteBufferPos].dwFlags = 0;
		if ((LeftVolume & RightVolume) != 0xFFFF)
			ChangeBufferVolume(WriteBufferPos);
		::waveOutPrepareHeader(hWaveOut,WaveHdr + WriteBufferPos,sizeof(WAVEHDR));
		::waveOutWrite(hWaveOut,WaveHdr + WriteBufferPos,sizeof(WAVEHDR));
		WriteBufferPos = (WriteBufferPos + 1) % BufferNum;
		LoadEnd = size < BufferSize;
		if (Stream->isEnd())
			return 2;
		return 0;
	}
	return 1;
}


bool waveOutStreamPlayer::SetBufferVolume(unsigned long left,unsigned long right)
{
//	Synchronizer sync(CritSect);
	LeftVolume = min(left,0x10000);
	RightVolume = min(right,0x10000);
	return true;
}

void waveOutStreamPlayer::ChangeBufferVolume(int bufIndex)
{
	double left = (double)LeftVolume / (double)0x10000;
	double right = (double)RightVolume / (double)0x10000;
	long samples = WaveHdr[bufIndex].dwBufferLength / PlayFormat.sample_size();
	long i;
	switch (PlayFormat.bit)
	{
	case 8: {
		BYTE *data = reinterpret_cast<BYTE *>(WaveHdr[bufIndex].lpData);
		switch (PlayFormat.channel)
		{
		case 1:
			for (i = 0;i < samples;i++,data++)
				*data = static_cast<BYTE>((static_cast<int>(*data) - 128) * left + 128);
			break;
		case 2:
			for (i = 0;i < samples;i++)
			{
				data[0] = static_cast<BYTE>((static_cast<int>(data[0]) - 128) * left + 128);
				data[1] = static_cast<BYTE>((static_cast<int>(data[1]) - 128) * right + 128);
				data += 2;
			}
			break;
		}
		} break;
	case 16: {
		short *data = reinterpret_cast<short *>(WaveHdr[bufIndex].lpData);
		switch (PlayFormat.channel)
		{
		case 1:
			for (i = 0;i < samples;i++,data++)
				*data = static_cast<short>(*data * left);
			break;
		case 2:
			for (i = 0;i < samples;i++)
			{
				data[0] = static_cast<short>(data[0] * left);
				data[1] = static_cast<short>(data[1] * right);
				data += 2;
			}
			break;
		}
		} break;
	}
}

/*=============================================================================
waveOutExVolumePlayer
=============================================================================*/
bool waveOutExVolumeStreamPlayer::Initialize
	(int device,const Format &format,long bufsize,int bufnum)
{
	if (!base_Initialize(device,format,bufsize,bufnum))
		return false;

	if (OriginalDataBuffer)
	{
		operator delete(OriginalDataBuffer);
		OriginalDataBuffer = 0;
	}
	OriginalDataBuffer = static_cast<void **>(operator new((BufferSize + sizeof(void *)) * BufferNum));
	if (!OriginalDataBuffer)
	{
		base_Terminalize();
		return false;
	}
	for (int i = 0;i < bufnum;i++)
	{
		OriginalDataBuffer[i] = reinterpret_cast<char *>(OriginalDataBuffer + BufferNum) + (BufferSize * i);
	}

	return true;
}

bool waveOutExVolumeStreamPlayer::Initialize
	(int device,I_InputStream &stream,long bufsize,int bufnum)
{
	if (!Initialize(device,stream.GetFormat(),bufsize,bufnum))
		return false;
	if (!SetSupplier(&stream))
	{
		Terminalize();
		return false;
	}
	return true;
}

void waveOutExVolumeStreamPlayer::Terminalize(void)
{
	base_Terminalize();
	if (OriginalDataBuffer)
	{
		operator delete(OriginalDataBuffer);
		OriginalDataBuffer = 0;
	}
}

int waveOutExVolumeStreamPlayer::LoadStream(void)
{
	if (WaveHdr[WriteBufferPos].dwFlags & WHDR_DONE)
	{
		unsigned long size = Stream->Read(OriginalDataBuffer[WriteBufferPos],BufferSize);
		WaveHdr[WriteBufferPos].dwBufferLength = size;
		WaveHdr[WriteBufferPos].dwFlags = 0;
//		memcpy(WaveHdr[WriteBufferPos].lpData,OriginalDataBuffer[WriteBufferPos],size);
//		if (Volume < 0xFFFF)
		{
			ChangeBufferVolume(WriteBufferPos);
		}
		::waveOutPrepareHeader(hWaveOut,WaveHdr + WriteBufferPos,sizeof(WAVEHDR));
		::waveOutWrite(hWaveOut,WaveHdr + WriteBufferPos,sizeof(WAVEHDR));
		WriteBufferPos = (WriteBufferPos + 1) % BufferNum;
		if (Stream->isEnd())
		{
			LoadEnd = true;
			return 2;
		}
		LoadEnd = false;
		return 0;
	}
	return 1;
}


//{[ύX@㏸̏ꍇAʂɃIWĩf[^ێȂƃ_
bool waveOutExVolumeStreamPlayer::SetBufferVolume
	(unsigned long left,unsigned long right)
{
	LeftVolume = min(left,0x10000);
	RightVolume = min(right,0x10000);
	if (!(ThreadState == TS_STANDBY && WriteBufferPos == 0))
	{
		DWORD playpos = WriteBufferPos;
		int count = 0;
//ݍĐʒu
		while (WaveHdr[playpos].dwFlags & WHDR_DONE)
		{
			playpos = (playpos + 1) % BufferNum;
			if (count++ > BufferNum)
				return false;	//NȂ͂[v΍
		}
		unsigned playing = playpos;

//Đł
		ChangeBufferVolume(playpos);
		playpos = (playpos + 1) % BufferNum;
		while (!(WaveHdr[playpos].dwFlags & WHDR_DONE))
		{
//waveOutWritêł̂܂܏
			ChangeBufferVolume(playpos);
			playpos = (playpos + 1) % BufferNum;
			if (playing == playpos)
				break;
		}
	}
	return true;
}//*/


void waveOutExVolumeStreamPlayer::ChangeBufferVolume(int bufIndex)
{
//	unsigned long left = LeftVolume + (LeftVolume == 0xFFFF);
//	unsigned long right = RightVolume + (RightVolume == 0xFFFF);
	double left = (double)LeftVolume / (double)0x10000;
	double right = (double)RightVolume / (double)0x10000;
	long samples = WaveHdr[bufIndex].dwBufferLength / PlayFormat.sample_size();
	long i;
	switch (PlayFormat.bit)
	{
	case 8:
		{
		BYTE *dest = reinterpret_cast<BYTE *>(WaveHdr[bufIndex].lpData);
		BYTE *src = static_cast<BYTE *>(OriginalDataBuffer[bufIndex]);
		switch (PlayFormat.channel)
		{
		case 1:
			for (i = 0;i < samples;i++)
			{
				*dest = static_cast<BYTE>((static_cast<int>(*src) - 128) * left + 128);
				dest++;src++;
			}
			break;
		case 2:
			for (i = 0;i < samples;i++)
			{
				dest[0] = static_cast<BYTE>((static_cast<int>(src[0]) - 128) * left + 128);
				dest[1] = static_cast<BYTE>((static_cast<int>(src[1]) - 128) * right + 128);
				dest += 2;src += 2;
			}
			break;
		}
		break;
		}
	case 16:
		{
		short *dest = reinterpret_cast<short *>(WaveHdr[bufIndex].lpData);
		short *src = static_cast<short *>(OriginalDataBuffer[bufIndex]);
		switch (PlayFormat.channel)
		{
		case 1:
			for (i = 0;i < samples;i++)
			{
				*dest = static_cast<short>(*src * left);
				dest++;src++;
			}
			break;
		case 2:
			for (i = 0;i < samples;i++)
			{
				dest[0] = static_cast<short>(src[0] * left);
				dest[1] = static_cast<short>(src[1] * right);
				dest += 2;src += 2;
			}
			break;
		}
		break;
		}
	}
}


}//namespace Sound
}//namespace Juna
