
#include "../../DSCaptureDriver.h"
#include "../../DSCaptureRecorder.h"

#include "../ExtensibleFormat.h"

namespace Juna {
namespace Sound {

DSCaptureRecordDriver::DSCaptureRecordDriver(void)
	: list(NULL),wlist(NULL),IndexNum(-1),hDSdll(NULL) {}

DSCaptureRecordDriver::~DSCaptureRecordDriver() {DeleteDeviceList();}

I_StreamRecorder *DSCaptureRecordDriver::CreateRecorder
	(int iDevice,const Format &iFormat,long iBufsize,int iBufnum)
{
	DirectSoundCaptureRecorder *r;
	r = new DirectSoundCaptureRecorder();
	if (r && !r->Initialize(iDevice,iFormat,iBufsize,iBufnum))
	{
		delete r;
		return 0;
	}
	return r;
}

bool DSCaptureRecordDriver::DestroyRecorder(I_StreamRecorder *recorder)
{
	if (recorder->GetDriver() != this)
		return false;;
	delete static_cast<DirectSoundCaptureRecorder *>(recorder);
	return true;
}

int DSCaptureRecordDriver::GetDeviceNum(void)
{
	if (IndexNum < 0) InitDeviceNum();
	return IndexNum;
}

const DSCaptDeviceData *DSCaptureRecordDriver::GetDSCDeviceData(int index)
{
	if (list == NULL) UpdateDeviceList();
	if (index < 0 || index >= IndexNum)
		return NULL;
	return &list[index];
}

const DSCaptDeviceDataW *DSCaptureRecordDriver::GetDSCDeviceDataW(int index)
{
	if (wlist == NULL) UpdateDeviceListW();
	if (index < 0 || index >= IndexNum)
		return NULL;
	return &wlist[index];
}

const GUID *DSCaptureRecordDriver::GetDSCDeviceGUID(int index)
{
	if (wlist)
	{
		const DSCaptDeviceDataW *dd = GetDSCDeviceDataW(index);
		if (dd == NULL)
			return NULL;
		return &dd->guid;
	}
	else
	{
		const DSCaptDeviceData *dd = GetDSCDeviceData(index);
		if (dd == NULL)
			return NULL;
		return &dd->guid;
	}
}


HINSTANCE DSCaptureRecordDriver::LoadDll(void)
{
	if (hDSdll == NULL) hDSdll= LoadLibraryA("dsound");
	return hDSdll;
}
void DSCaptureRecordDriver::FreeDll(void)
{
	if (hDSdll) FreeLibrary(hDSdll);
	hDSdll = NULL;
}

const DSCaptureRecordDriver::typeDSCEnumA DSCaptureRecordDriver::GetEnumProcA(void)
{
	if (LoadDll() == NULL)
		return NULL;
	const typeDSCEnumA DSCEnumProc =
		(typeDSCEnumA)GetProcAddress(hDSdll,"DirectSoundCaptureEnumerateA");
	if (DSCEnumProc == NULL)
		FreeDll();
	return DSCEnumProc;
}

const DSCaptureRecordDriver::typeDSCEnumW DSCaptureRecordDriver::GetEnumProcW(void)
{
	if (LoadDll() == NULL)
		return NULL;
	const typeDSCEnumW DSCEnumProc =
		(typeDSCEnumW)GetProcAddress(hDSdll,"DirectSoundCaptureEnumerateW");
	if (DSCEnumProc == NULL)
		FreeDll();
	return DSCEnumProc;
}


void DSCaptureRecordDriver::InitDeviceNum(void)
{
	struct local {	//ȂƂȂ[J֐
		static BOOL CALLBACK DSCEnumCallback(LPGUID,LPCSTR,LPCSTR,LPVOID lpContext)
		{(*static_cast<int *>(lpContext))++;return TRUE;}
	};

	IndexNum = -1;
	const typeDSCEnumA DSCEnumProc = GetEnumProcA();
	if (DSCEnumProc == NULL)
		return;
	IndexNum = 0;
	DSCEnumProc(local::DSCEnumCallback,&IndexNum);
	return;
}


void DSCaptureRecordDriver::UpdateDeviceList(void)
{
	DeleteDeviceList();
	InitDeviceNum();
	if (IndexNum <= 0)
		return;

	//pm
	unsigned long needsize = 0;
	struct getsize
	{
		static BOOL CALLBACK DSCEnumCallback(LPGUID,LPCSTR lpcstrDesc,LPCSTR lpcstrModule,LPVOID lpContext)
		{
			unsigned long &size = *static_cast<unsigned long *>(lpContext);
			size += lstrlenA(lpcstrDesc) + lstrlenA(lpcstrModule) + 2;
			return TRUE;
		}
	};
	const typeDSCEnumA DSCEnumProc = GetEnumProcA();
	if (DSCEnumProc == NULL) return;
	DSCEnumProc(getsize::DSCEnumCallback,&needsize);
	list = static_cast<DSCaptDeviceData *>(operator new(sizeof(DSCaptDeviceData) * IndexNum + needsize));

	struct local
	{
		struct Data {	//[Jm͌ȂHdȂ
			DSCaptDeviceData *list;
			int max,i;
			char *buf;
		};
		static BOOL CALLBACK DSCEnumCallback(LPGUID lpGuid,LPCSTR lpcstrDescription,LPCSTR lpcstrModule,LPVOID lpContext)
		{
			Data *data = static_cast<Data *>(lpContext);
			if (data->i >= data->max)
				return FALSE;
			if (lpGuid)
				data->list[data->i].guid = *lpGuid;
			else
				data->list[data->i].guid = GUID_NULL;

			data->list[data->i].name = data->buf;
			lstrcpyA(data->buf,lpcstrDescription);
			data->buf += lstrlenA(data->buf) + 1;

			data->list[data->i].Module = data->buf;
			lstrcpyA(data->buf,lpcstrModule);
			data->buf += lstrlenA(data->buf) + 1;

			data->i++;
			return TRUE;
		}
	};//struct local

	local::Data data = {list,IndexNum,0,reinterpret_cast<char *>(list + IndexNum)};
	DSCEnumProc(local::DSCEnumCallback,&data);

	return;
}

void DSCaptureRecordDriver::UpdateDeviceListW(void)
{
	DeleteDeviceList();
	InitDeviceNum();
	if (IndexNum <= 0)
		return;

	//pm
	unsigned long needsize = 0;
	struct getsize
	{
		static BOOL CALLBACK DSCEnumCallback(LPGUID,LPCWSTR lpcstrDesc,LPCWSTR lpcstrModule,LPVOID lpContext)
		{
			unsigned long &size = *static_cast<unsigned long *>(lpContext);
			size += (lstrlenW(lpcstrDesc) + lstrlenW(lpcstrModule) + 2) * sizeof(unsigned short);
			return TRUE;
		}
	};
	const typeDSCEnumW DSCEnumProc = GetEnumProcW();
	if (DSCEnumProc == NULL) return;
	DSCEnumProc(getsize::DSCEnumCallback,&needsize);
	wlist = static_cast<DSCaptDeviceDataW *>(operator new(sizeof(DSCaptDeviceDataW) * IndexNum + needsize));

	struct local
	{
		struct Data {	//[Jm͌ȂHdȂ
			DSCaptDeviceDataW *wlist;
			int max,i;
			unsigned short *buf;
		};
		static BOOL CALLBACK DSCEnumCallback(LPGUID lpGuid,LPCWSTR lpcstrDescription,LPCWSTR lpcstrModule,LPVOID lpContext)
		{
			Data *data = static_cast<Data *>(lpContext);
			if (data->i >= data->max)
				return FALSE;
			if (lpGuid)
				data->wlist[data->i].guid = *lpGuid;
			else
				data->wlist[data->i].guid = GUID_NULL;

			data->wlist[data->i].name = data->buf;
			lstrcpyW(data->buf,lpcstrDescription);
			data->buf += lstrlenW(data->buf) + 1;

			data->wlist[data->i].Module = data->buf;
			lstrcpyW(data->buf,lpcstrModule);
			data->buf += lstrlenW(data->buf) + 1;

			data->i++;
			return TRUE;
		}
	};//struct local

	local::Data data = {wlist,IndexNum,0,reinterpret_cast<unsigned short *>(wlist + IndexNum)};
	DSCEnumProc(local::DSCEnumCallback,&data);

	return;
}

void DSCaptureRecordDriver::DeleteDeviceList(void)
{
	if (list) operator delete(list);
	list = NULL;
	if (wlist) operator delete(wlist);
	wlist = NULL;
	FreeDll();
	IndexNum = -1;
}

bool DSCaptureRecordDriver::TestOpenDevice(int index,const Format &format)
{
	if ((format.bit == 8 && format.quantization == QT_SIGNED) ||
		(format.bit != 8 && format.quantization == QT_UNSIGNED))
		return false;
	LPCGUID guid = GetDSCDeviceGUID(index);
	if (guid == NULL)
		return false;
	if (IsEqualGUID(*guid,GUID_NULL))
		guid = NULL;
	const typeDSCCreate DSCCreate = GetDSCCreateProc();
	if (DSCCreate == NULL)
		return false;

	LPDIRECTSOUNDCAPTURE lpDSCObject;
	if FAILED(DSCCreate(guid,&lpDSCObject,NULL))
		return false;

	WAVEFORMATEX wfx;
	SetWaveFormatEx(wfx,format);
	DSCBUFFERDESC dscbd = {sizeof(dscbd),0,format.sample_size() * 1,0,&wfx};

	LPDIRECTSOUNDCAPTUREBUFFER lpBuffer;
	if (format.bit > 16 || format.channel > 2 ||
		FAILED(lpDSCObject->CreateCaptureBuffer(&dscbd,&lpBuffer,NULL)))
	{	//ExtensibleFormat
		const int extsize = sizeof(WAVEFORMATEX) + 22;
		BYTE buf[extsize];
		WAVEFORMATEX *wfext = reinterpret_cast<WAVEFORMATEX *>(buf);
		memcpy(wfext,&wfx,sizeof(WAVEFORMATEX));
		wfext->cbSize = extsize - sizeof(WAVEFORMATEX);
		if (!MakeExtensibleFormat(wfext)) {
			lpDSCObject->Release();
			return false;
		}
		dscbd.lpwfxFormat = wfext;
		if FAILED(lpDSCObject->CreateCaptureBuffer(&dscbd,&lpBuffer,NULL)) {
			lpDSCObject->Release();
			return false;
		}
	}
	lpBuffer->Release();
	lpDSCObject->Release();
	return true;
}

const DSCaptureRecordDriver::typeDSCCreate DSCaptureRecordDriver::GetDSCCreateProc(void)
{
	if (LoadDll() == NULL)
		return NULL;
	return (typeDSCCreate)GetProcAddress(hDSdll,"DirectSoundCaptureCreate");
}



DSCaptureRecordDriver *DSCaptureRecordDriver::GetInstance(void)
{
	static DSCaptureRecordDriver Maker;
	return &Maker;
}

I_DSCaptureRecordDriver *GetDSCaptureRecordDriver(void)
{
	return DSCaptureRecordDriver::GetInstance();
}


}//namespace Sound
}//namespace Juna
