
#include "../../DSCaptureRecorder.h"
#include "../../DSCaptureDriver.h"

#include "../ExtensibleFormat.h"

#include <process.h>

#include "DrainReceiver.h"

namespace Juna {
namespace Sound {


namespace {
class Synchronizer
{
public:
	Synchronizer(CRITICAL_SECTION &locker) : Locker(locker)
	{::EnterCriticalSection(&Locker);}
	~Synchronizer(void)
	{::LeaveCriticalSection(&Locker);}
private:
	CRITICAL_SECTION &Locker;
};
}//namespace ""


DirectSoundCaptureRecorder::DirectSoundCaptureRecorder(void)
  : lpDirectSoundCaptureObject(NULL),
	lpDSCBuffer(NULL),
	RecordFormat(),
	BufferSize(0),BufferNum(0),ReadBufferPos(0),
	Receiver(&RecorderDrainStream),
	hThread(NULL),ThreadState(TS_EXIT)
{hNotifyEvent[0]=NULL;hNotifyEvent[1]=NULL;}


unsigned __stdcall DirectSoundCaptureRecorder::RecordThread(LPVOID DSCRecorder)
{
	DirectSoundCaptureRecorder * const this_ =
			static_cast<DirectSoundCaptureRecorder *>(DSCRecorder);

	DWORD readpos,captured;
	DWORD bufsize;
	while (1)
	{
		::WaitForMultipleObjects(2,this_->hNotifyEvent,FALSE,INFINITE);

		Synchronizer sync(this_->CritSect);
		if (this_->ThreadState == TS_EXIT)
			break;
		if (this_->ThreadState == TS_STANDBY)
			continue;

		bufsize = this_->BufferSize * this_->BufferNum;
		captured = readpos = this_->ReadBufferPos * this_->BufferSize;
		this_->lpDSCBuffer->GetCurrentPosition(NULL,&captured);
		captured = (captured - readpos) + bufsize * (captured < readpos);
		while (captured >= static_cast<DWORD>(this_->BufferSize))
		{
			void *lock;
			DWORD lsize;
			this_->lpDSCBuffer->Lock(readpos,this_->BufferSize,&lock,&lsize,NULL,NULL,0);
			this_->Receiver->Write(lock,lsize);
			this_->lpDSCBuffer->Unlock(lock,lsize,NULL,NULL);
			this_->ReadBufferPos = (this_->ReadBufferPos + 1) % this_->BufferNum;

			captured = readpos = this_->ReadBufferPos * this_->BufferSize;
			this_->lpDSCBuffer->GetCurrentPosition(NULL,&captured);
			captured = (captured - readpos) + bufsize * (captured < readpos);
		}
		if (::WaitForSingleObject(this_->hNotifyEvent[1],0) == WAIT_OBJECT_0)
		{	//^~̂Ŏc̕S
			ResetEvent(this_->hNotifyEvent[1]);
			captured = readpos = this_->ReadBufferPos * this_->BufferSize;
			this_->lpDSCBuffer->GetCurrentPosition(NULL,&captured);
			captured = (captured - readpos) + bufsize * (captured < readpos);
			void *lock1,*lock2;
			DWORD lsize1,lsize2;
			this_->lpDSCBuffer->Lock(readpos,captured,&lock1,&lsize1,&lock2,&lsize2,0);
			this_->Receiver->Write(lock1,lsize1);
			if (lock2)
				this_->Receiver->Write(lock2,lsize2);
			this_->lpDSCBuffer->Unlock(lock1,lsize1,lock2,lsize2);

			//Jnʒuŏɖ߂
			this_->ReadBufferPos = 0;
			this_->lpDSCBuffer->Start(0);
			::WaitForSingleObject(this_->hNotifyEvent[1],INFINITE);
			::ResetEvent(this_->hNotifyEvent[1]);
			this_->ThreadState = this_->TS_STANDBY;
		}
	}
	return 0;
}

bool DirectSoundCaptureRecorder::Start(void)
{
	if (lpDSCBuffer == NULL)
		return false;
	Synchronizer sync(CritSect);
	if (ThreadState & TS_RECORD)
		return true;
	ThreadState = TS_RECORD;
	return (lpDSCBuffer->Start(DSCBSTART_LOOPING) == DS_OK);
}

bool DirectSoundCaptureRecorder::Stop(void)
{
	if (lpDSCBuffer == NULL)
		return false;
	Synchronizer sync(CritSect);
	if (ThreadState == TS_RECORD)
	{
		ThreadState = TS_RECORDEND;
		lpDSCBuffer->Stop();
		Sleep(1);
	}
	return true;
}

bool DirectSoundCaptureRecorder::isRecording(void) const
{
	if (lpDSCBuffer == NULL)
		return false;
//	Synchronizer sync(CritSect);
	return (ThreadState & TS_RECORD) != 0;
}


bool DirectSoundCaptureRecorder::SetReceiver
	(I_OutputStream *receiver)
{
	if (lpDSCBuffer == NULL)
		return false;
	Stop();
	Synchronizer sync(CritSect);

	if (receiver == Receiver)
		return true;
	if (receiver)
	{
		if (receiver->GetFormat() != RecordFormat)
			return false;
	}
	else
		receiver = &RecorderDrainStream;

	receiver->AddRef();
	Receiver->Release();
	Receiver = receiver;
	return true;
}

const I_RecordDriver *DirectSoundCaptureRecorder::GetDriver(void) const
{return DSCaptureRecordDriver::GetInstance();}


I_StreamRecorder *DirectSoundCaptureRecorder::Initialize
	(int device,const Format &format,long bufsize,int bufnum)
{
	HRESULT hr;
	if (lpDSCBuffer)
		Finalize();
	if ((format.bit == 8 && format.quantization == QT_SIGNED) ||
		(format.bit != 8 && format.quantization == QT_UNSIGNED))
		return 0;
	if (bufnum == 0)
		return 0;

	LPCGUID guid = DSCaptureRecordDriver::GetInstance()->GetDSCDeviceGUID(device);
	if (guid == NULL)
		return 0;
	if (::IsEqualGUID(*guid,GUID_NULL))
		guid = NULL;

	const DSCaptureRecordDriver::typeDSCCreate DSCCreate =
		DSCaptureRecordDriver::GetInstance()->GetDSCCreateProc();
	if FAILED(DSCCreate(guid,&lpDirectSoundCaptureObject,NULL))
	{
		lpDirectSoundCaptureObject = NULL;
		return 0;
	}

	if (bufsize == 0)	//ftHgł0w肳ƎIɌ
		bufsize = format.frequency * format.sample_size() * 1;
	if (bufsize % format.sample_size())	//TvTCYɍ킹
		bufsize -= bufsize % format.sample_size();
	BufferSize = bufsize;
	BufferNum = bufnum;

	WAVEFORMATEX wfx;
	SetWaveFormatEx(wfx,format);

	DSCBUFFERDESC dscbd = {sizeof(dscbd),0,bufsize * bufnum,0,&wfx};

	LPDIRECTSOUNDCAPTUREBUFFER lpBuffer;
	if (format.bit > 16 || format.channel > 2 ||
		FAILED(lpDirectSoundCaptureObject->CreateCaptureBuffer(&dscbd,&lpBuffer,NULL)))
	{	//ExtensibleFormat
		const int extsize = sizeof(WAVEFORMATEX) + 22;
		BYTE buf[extsize];
		WAVEFORMATEX *wfext = reinterpret_cast<WAVEFORMATEX *>(buf);
		::memcpy(wfext,&wfx,sizeof(WAVEFORMATEX));
		wfext->cbSize = extsize - sizeof(WAVEFORMATEX);
		if (!MakeExtensibleFormat(wfext))
			goto ERR;
		dscbd.lpwfxFormat = wfext;
		if FAILED(lpDirectSoundCaptureObject->CreateCaptureBuffer(&dscbd,&lpBuffer,NULL))
			goto ERR;
	}
	lpDSCBuffer = lpBuffer;
//NotifyEventݒ
	hNotifyEvent[0] = ::CreateEvent(NULL,FALSE,FALSE,NULL);
	if (hNotifyEvent[0] == NULL)
		goto ERR;
	hNotifyEvent[1] = ::CreateEvent(NULL,TRUE,FALSE,NULL);
	if (hNotifyEvent[1] == NULL)
		goto ERR;
	{	//goto ERR;̏łȂG[@OgƂ
		DSBPOSITIONNOTIFY *dsbpn = new DSBPOSITIONNOTIFY[bufnum + 1];
		for (int i = 0;i < bufnum;i++)
		{
			dsbpn[i].dwOffset = bufsize * i;
			dsbpn[i].hEventNotify = hNotifyEvent[0];
		}
		dsbpn[bufnum].dwOffset = DSBPN_OFFSETSTOP;
		dsbpn[bufnum].hEventNotify = hNotifyEvent[1];

		LPDIRECTSOUNDNOTIFY lpdsNotify = NULL;
		if FAILED(lpDSCBuffer->QueryInterface(IID_IDirectSoundNotify,(VOID **)&lpdsNotify))
		{
			delete[] dsbpn;
			goto ERR;
		}
		hr = lpdsNotify->SetNotificationPositions(bufnum + 1,dsbpn);
		lpdsNotify->Release();	//ݒ肳ł΂̃C^[tFCX͂
		delete[] dsbpn;
		if FAILED(hr)
			goto ERR;
	}

	ThreadState = TS_STANDBY;
	::InitializeCriticalSection(&CritSect);
	unsigned threadID;
	hThread = (HANDLE)::_beginthreadex(NULL,0,RecordThread,this,0,&threadID);
	if (hThread == NULL) {
		::DeleteCriticalSection(&CritSect);
		goto ERR;
	}
	::SetThreadPriority(hThread,THREAD_PRIORITY_ABOVE_NORMAL);

	RecordFormat = format;
	return this;

ERR:;
	Finalize();
	return 0;
}

void DirectSoundCaptureRecorder::Finalize(void)
{
	SetReceiver(0);
	if (hThread)
	{
//ʂɌJԂKvȂǁBقƂ͈莞ԌTerminateThreadŋIƂ
		do
		{
			ThreadState = TS_EXIT;		//XbhItOZbg
			::SetEvent(hNotifyEvent[0]);
//Î҂AIĂȂx
		} while (WaitForSingleObject(hThread,10) == WAIT_TIMEOUT);
		hThread = NULL;
		::DeleteCriticalSection(&CritSect);
	}
	if (hNotifyEvent[0])
	{
		::CloseHandle(hNotifyEvent[0]);
		hNotifyEvent[0] = NULL;
	}
	if (hNotifyEvent[1])
	{
		::CloseHandle(hNotifyEvent[1]);
		hNotifyEvent[1] = NULL;
	}
	if (lpDSCBuffer)
	{
		lpDSCBuffer->Release();
		lpDSCBuffer = NULL;
	}
	if (lpDirectSoundCaptureObject)
	{
		lpDirectSoundCaptureObject->Release();
		lpDirectSoundCaptureObject = NULL;
	}
	RecordFormat.reset();
	BufferSize = BufferNum = 0;
}



}//namespace Sound
}//namespace Juna
