
#include "../../waveInDriver.h"

#include "../../waveInRecorder.h"

#include "../ExtensibleFormat.h"

namespace Juna {
namespace Sound
{


waveInRecordDriver::waveInRecordDriver(void) : list(NULL),wlist(NULL) {}

waveInRecordDriver::~waveInRecordDriver() {DeleteDeviceList();}

I_StreamRecorder *waveInRecordDriver::CreateRecorder
		(int iDevice,const Format &iFormat,long iBufsize,int iBufnum)
{
	waveInRecorder *r;
	r = new waveInRecorder();
	if (r && !r->Initialize(iDevice,iFormat,iBufsize,iBufnum))
	{
		delete r;
		return 0;
	}
	return r;
}
bool waveInRecordDriver::DestroyRecorder(I_StreamRecorder *recorder)
{
	if (recorder->GetDriver() != this)
		return false;;
	delete static_cast<waveInRecorder *>(recorder);
	return true;
}

void waveInRecordDriver::MakeDeviceList(void)
{
	DeleteDeviceList();
	int count = GetDeviceNum();
	if (count == 0) return;
	list = new waveInDeviceData[count];
	for (int i = 0;i < count;i++)
	{
		::waveInGetDevCapsA(i-1,&list[i].wicaps,sizeof(WAVEOUTCAPSA));
		list[i].ID = i - 1;
		list[i].name = list[i].wicaps.szPname;
	}
}

void waveInRecordDriver::MakeDeviceListW(void)
{
	DeleteDeviceList();
	int count = GetDeviceNum();
	if (count == 0) return;
	wlist = new waveInDeviceDataW[count];
	for (int i = 0;i < count;i++)
	{
		waveInGetDevCapsW(i-1,&wlist[i].wicaps,sizeof(WAVEOUTCAPSW));
		wlist[i].ID = i - 1;
		wlist[i].name = wlist[i].wicaps.szPname;
	}
}


void waveInRecordDriver::DeleteDeviceList(void)
{
	if (list) delete[] list;
	list = NULL;
	if (wlist) delete[] wlist;
	wlist = NULL;
}

int waveInRecordDriver::GetDeviceNum(void)
{
	int count = waveInGetNumDevs();
	return count + (count > 0);
}

const RecordDeviceData *waveInRecordDriver::GetDeviceData(int index)
{return GetWIDeviceData(index);}

const waveInDeviceData *waveInRecordDriver::GetWIDeviceData(int index)
{
	if (list == NULL) MakeDeviceList();
	if (index < 0 || index >= GetDeviceNum())
		return NULL;
	return &list[index];
}

const RecordDeviceDataW *waveInRecordDriver::GetDeviceDataW(int index)
{return GetWIDeviceDataW(index);}

const waveInDeviceDataW *waveInRecordDriver::GetWIDeviceDataW(int index)
{
	if (wlist == NULL) MakeDeviceListW();
	if (index < 0 || index >= GetDeviceNum())
		return NULL;
	return &wlist[index];
}

const UINT *waveInRecordDriver::GetWIDeviceID(int index)
{
	if (wlist)
	{
		const waveInDeviceDataW *dd = GetInstance()->GetWIDeviceDataW(index);
		if (dd == NULL)
			return NULL;
		return &dd->ID;
	}
	else
	{
		const waveInDeviceData *dd = GetInstance()->GetWIDeviceData(index);
		if (dd == NULL)
			return NULL;
		return &dd->ID;
	}
}


bool waveInRecordDriver::TestOpenDevice(int index,const Format &format)
{
	if ((format.bit == 8 && format.quantization == QT_SIGNED) ||
		(format.bit != 8 && format.quantization == QT_UNSIGNED))
		return false;

	const UINT *id = GetWIDeviceID(index);
	if (id == NULL)
		return false;

	WAVEFORMATEX wfx;
	SetWaveFormatEx(wfx,format);
	HWAVEIN hwi;
	DWORD openflag = CALLBACK_NULL  | WAVE_FORMAT_DIRECT;
	if (format.bit > 16 || format.channel > 2 ||
		waveInOpen(&hwi,*id,&wfx,0,0,openflag) != MMSYSERR_NOERROR)
	{
		const int extsize = sizeof(WAVEFORMATEX) + 22;
		BYTE buf[extsize];
		WAVEFORMATEX *wfext = reinterpret_cast<WAVEFORMATEX *>(buf);
		memcpy(wfext,&wfx,sizeof(WAVEFORMATEX));
		wfext->cbSize = extsize - sizeof(WAVEFORMATEX);
		if (!MakeExtensibleFormat(wfext))
			return false;
		if (waveInOpen(&hwi,*id,wfext,0,0,openflag) != MMSYSERR_NOERROR)
			return false;
	}
	waveInClose(hwi);
	return true;
}




waveInRecordDriver *waveInRecordDriver::GetInstance(void)
{
	static waveInRecordDriver Driver;
	return &Driver;
}

I_waveInRecordDriver *GetWaveInRecordDriver(void)
{
	return waveInRecordDriver::GetInstance();
}


}//namespace Sound
}//namespace Juna
