
#include "../../waveInRecorder.h"
#include "../../waveInDriver.h"

#include "../ExtensibleFormat.h"

#include <process.h>

#pragma comment(lib, "winmm.lib")

#include "DrainReceiver.h"

namespace Juna {
namespace Sound {


namespace {
class Synchronizer
{
public:
	Synchronizer(CRITICAL_SECTION &locker) : Locker(locker)
	{::EnterCriticalSection(&Locker);}
	~Synchronizer(void)
	{::LeaveCriticalSection(&Locker);}
private:
	CRITICAL_SECTION &Locker;
};

}//namespace ""

waveInRecorder::waveInRecorder(void)
  : hWaveIn(NULL),RecordFormat(),
	WaveHdr(NULL),BufferSize(0),BufferNum(0),ReadBufferPos(0),
	hCallbackEvent(NULL),hThread(NULL),ThreadState(TS_EXIT),
	Receiver(&RecorderDrainStream)
{}

unsigned __stdcall waveInRecorder::RecordThread(void *wiRecorder)
{
	waveInRecorder * const this_ = (waveInRecorder *)wiRecorder;

	while (1)
	{
		::WaitForSingleObject(this_->hCallbackEvent,INFINITE);

		Synchronizer sync(this_->CritSect);
		switch (this_->ThreadState)
		{
		case TS_EXIT:
			return 0;
		case TS_RECORD:
			while (this_->WaveHdr[this_->ReadBufferPos].dwFlags & WHDR_DONE)
			{
				this_->Receiver->Write(this_->WaveHdr[this_->ReadBufferPos].lpData,
									   this_->WaveHdr[this_->ReadBufferPos].dwBytesRecorded);
				::waveInPrepareHeader(this_->hWaveIn,this_->WaveHdr + this_->ReadBufferPos,sizeof(WAVEHDR));
				::waveInAddBuffer(this_->hWaveIn,this_->WaveHdr + this_->ReadBufferPos,sizeof(WAVEHDR));
				this_->ReadBufferPos = (this_->ReadBufferPos + 1) % this_->BufferNum;
			}
			break;
		case TS_RECORDEND:
//			waveInReset(this_->hWaveIn);
//obt@ɎcĂf[^̏
			while (this_->WaveHdr[this_->ReadBufferPos].dwFlags & WHDR_DONE)
			{
//^ɎgĂȂobt@͖
				if (this_->WaveHdr[this_->ReadBufferPos].dwBytesRecorded > 0)
				{
					this_->Receiver->Write(this_->WaveHdr[this_->ReadBufferPos].lpData,
										   this_->WaveHdr[this_->ReadBufferPos].dwBytesRecorded);
				}
				::waveInPrepareHeader(this_->hWaveIn,this_->WaveHdr + this_->ReadBufferPos,sizeof(WAVEHDR));
				::waveInAddBuffer(this_->hWaveIn,this_->WaveHdr + this_->ReadBufferPos,sizeof(WAVEHDR));
				this_->ReadBufferPos = (this_->ReadBufferPos + 1) % this_->BufferNum;
			}
			this_->ThreadState = TS_STANDBY;
			break;
		}
	}
//BR[h
	return 0;
}



bool waveInRecorder::Start(void)
{
	if (hWaveIn == NULL)
		return false;
	Synchronizer sync(CritSect);

	if (ThreadState & TS_RECORD)
		return true;
	ThreadState = TS_RECORD;
	return (::waveInStart(hWaveIn) == MMSYSERR_NOERROR);
}

bool waveInRecorder::Stop(void)
{
	if (hWaveIn == NULL)
		return false;
	Synchronizer sync(CritSect);
	if (ThreadState == TS_RECORD)
	{
		::waveInReset(hWaveIn);
		ThreadState = TS_RECORDEND;
		::SetEvent(hCallbackEvent);
		::Sleep(1);
	}
	return true;
}

bool waveInRecorder::isRecording(void) const
{
	if (hWaveIn == NULL)
		return false;
//	Synchronizer sync(CritSect);
	return (ThreadState & TS_RECORD) != 0;
}


bool waveInRecorder::SetReceiver(I_OutputStream *receiver)
{
	if (hWaveIn == NULL)
		return false;
	Stop();
	Synchronizer sync(CritSect);

	if (receiver == Receiver)
		return true;
	if (receiver)
	{
		if (receiver->GetFormat() != RecordFormat)
			return false;
	}
	else
		receiver = &RecorderDrainStream;

	receiver->AddRef();
	Receiver->Release();
	Receiver = receiver;
	return true;
}

const I_RecordDriver *waveInRecorder::GetDriver(void) const
{return waveInRecordDriver::GetInstance();}


bool waveInRecorder::SetupDevice
	(int device,const Format &format,long bufsize,int bufnum)
{
	if ((format.bit == 8 && format.quantization == QT_SIGNED) ||
		(format.bit != 8 && format.quantization == QT_UNSIGNED))
		return false;

	const UINT *id = waveInRecordDriver::GetInstance()->GetWIDeviceID(device);
	if (id == NULL)
		return false;

	hCallbackEvent = ::CreateEvent(NULL,FALSE,FALSE,NULL);
	if (hCallbackEvent == NULL)
		return false;

	WAVEFORMATEX wfx;
	SetWaveFormatEx(wfx,format);
	DWORD openflag = CALLBACK_EVENT | WAVE_FORMAT_DIRECT;
//	if (deviceID == WAVE_MAPPER) openflag |= WAVE_MAPPED;
	HWAVEIN hwi;
	if (format.bit > 16 || format.channel > 2 ||
		::waveInOpen(&hwi,*id,&wfx,(DWORD)hCallbackEvent,(DWORD)this,openflag) != MMSYSERR_NOERROR)
	{
		const int extsize = sizeof(WAVEFORMATEX) + 22;
		BYTE buf[extsize];
		WAVEFORMATEX *wfext = reinterpret_cast<WAVEFORMATEX *>(buf);
		::memcpy(wfext,&wfx,sizeof(WAVEFORMATEX));
		wfext->cbSize = extsize - sizeof(WAVEFORMATEX);
		if (!MakeExtensibleFormat(wfext))
			goto ERR;
		if (::waveInOpen(&hwi,*id,wfext,(DWORD)hCallbackEvent,(DWORD)this,openflag) != MMSYSERR_NOERROR)
			goto ERR;
	}
	hWaveIn = hwi;

	if (bufsize == 0)	//ftHgł0w肳ƎIɌ
		bufsize = format.frequency * format.sample_size() * 1;
//TvTCYɍ킹
	if (bufsize % format.sample_size())
		bufsize -= bufsize % format.sample_size();
//Ȃ񂩃[mۂł̓G[łȂ`FbN
	if (bufnum == 0)
		goto ERR;
	BufferSize = bufsize;
	BufferNum = bufnum;

//ŏɈCɃm
	WaveHdr = static_cast<WAVEHDR *>(operator new((sizeof(WAVEHDR) + bufsize) * bufnum));

	int i;
	for (i = 0;i < bufnum;i++)
	{
		::memset(WaveHdr + i,0,sizeof(WAVEHDR));
		WaveHdr[i].lpData = ((LPSTR)(WaveHdr + bufnum)) + i * bufsize;
		WaveHdr[i].dwBufferLength = bufsize;

		if (::waveInPrepareHeader(hWaveIn,WaveHdr + i,sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
		{
			for (int j = 0;j < i;j++)
				::waveInUnprepareHeader(hWaveIn,WaveHdr + j,sizeof(WAVEHDR));
			operator delete (WaveHdr);
			WaveHdr = NULL;
			goto ERR;
		}
	}
	for (i = 0;i < bufnum;i++)
	{
		if (::waveInAddBuffer(hWaveIn,WaveHdr + i,sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
			goto ERR;
	}

	ReadBufferPos = 0;
	RecordFormat = format;

	return true;

ERR:;
	Finalize();
	return false;
}

I_StreamRecorder *waveInRecorder::Initialize
	(int device,const Format &format,long bufsize,int bufnum)
{
	if (hWaveIn != 0)
		Finalize();
	if (!SetupDevice(device,format,bufsize,bufnum))
		return 0;

	ThreadState = TS_STANDBY;
	::InitializeCriticalSection(&CritSect);
	unsigned threadID;
	hThread = (HANDLE)_beginthreadex(NULL,0,RecordThread,this,0,&threadID);
	if (hThread == NULL) {
		::DeleteCriticalSection(&CritSect);
		Finalize();
		return 0;
	}

	return this;
}


void waveInRecorder::Finalize(void)
{
	SetReceiver(0);
	if (hThread)
	{
		do
		{
			ThreadState = TS_EXIT;		//XbhItOZbg
			::SetEvent(hCallbackEvent);
//Î҂AIĂȂx
		} while (::WaitForSingleObject(hThread,10) == WAIT_TIMEOUT);
		::CloseHandle(hThread);
		hThread = NULL;
		::DeleteCriticalSection(&CritSect);
	}
	if (WaveHdr)
	{
		::waveInReset(hWaveIn);
		for (int i = 0;i < BufferNum;i++)
			::waveInUnprepareHeader(hWaveIn,WaveHdr + i,sizeof(WAVEHDR));
		operator delete (WaveHdr);
		WaveHdr = NULL;
	}
	if (hCallbackEvent) {
		::CloseHandle(hCallbackEvent);
		hCallbackEvent = NULL;
	}
	if (hWaveIn) {
		::waveInClose(hWaveIn);
		hWaveIn = NULL;
	}
	RecordFormat.reset();
	BufferSize = BufferNum = 0;
}


}//namespace Sound
}//namespace Juna
