
#ifndef JunaSound_waveOutStreamPlayer_H
#define JunaSound_waveOutStreamPlayer_H

#include "WindowsPlayer.h"
#include <windows.h>
#include <mmsystem.h>

namespace Juna {
namespace Sound {


class base_waveOutStreamPlayer : public I_waveOutStreamPlayer
{
protected:
	friend class waveOutPlayDriver;
	base_waveOutStreamPlayer(void);
	virtual ~base_waveOutStreamPlayer();
public:

	virtual bool Start(void);
	virtual bool Stop(void);
	virtual bool isPlaying(void) const;
//
	virtual Format GetFormat(void) const {return PlayFormat;}
	virtual const I_PlayDriver *GetDriver(void);


	virtual bool SetSupplier(I_InputStream *setSupplier);
	virtual long GetPlayTime(void);
	virtual void GetBufferSize(unsigned long &oBufsize,int &oBufnum) const
	{oBufsize = BufferSize;oBufnum = BufferNum;}

	virtual void SetPlayEndCallback(I_PlayEndCallback *setCallback);

	virtual int LoadStream(void) = 0;


	virtual bool SetVolume(unsigned long iLeft,unsigned long iRight)
	{
		if (VolumeFlag)
			return SetBufferVolume(iLeft,iRight);
		else
			return SetDeviceVolume(iLeft,iRight);
	}
	virtual bool GetVolume(unsigned long &left,unsigned long &right) const
	{
		if (VolumeFlag)
			return GetBufferVolume(left,right);
		else
			return GetDeviceVolume(left,right);
	}




	bool isInitialized(void) const {return hWaveOut != 0;}


	virtual bool &BufferVolumeFlag(void) {return VolumeFlag;}

//waveOut(Get/Set)Volume
	virtual bool SetDeviceVolume(unsigned long iLeft,unsigned long iRight);
	virtual bool GetDeviceVolume(unsigned long &oLeft,unsigned long &oRight) const;
//obt@Pʂ̎̓{[
	virtual bool SetBufferVolume(unsigned long iLeft,unsigned long iRight) = 0;
	virtual bool GetBufferVolume(unsigned long &oLeft,unsigned long &oRight) const
	{
		oRight = RightVolume;oLeft = LeftVolume;
		return hWaveOut != NULL;
	}
//obt@Pʂ̎̓{[

protected:
	bool base_Initialize(int device,const Format &iFormat,
						 long bufsize = 0,int bufnum = 3);
	void base_Terminalize(void);

protected:
	HWAVEOUT hWaveOut;		//ĐfoCX
	I_InputStream *Stream;	//ĐXg[

	WAVEHDR *WaveHdr;			//obt@
	unsigned long BufferSize;	//obt@TCY
	int BufferNum;				//obt@̐

	volatile int WriteBufferPos;	//ɏރobt@̈ʒu

	Format PlayFormat;

	enum THREAD_STATE{TS_EXIT = 0,TS_STANDBY = 1,TS_PLAY = 2,TS_PLAYEND = 3,TS_PAUSE = 4};
	volatile THREAD_STATE ThreadState;			//Đ
	HANDLE hDoneEvent;		//obt@ĐICxg
	HANDLE hThread;			//ĐICxg󂯂ăf[^[hXbh̃nh
	CRITICAL_SECTION CritSect;

	bool LoadEnd;			//TEhf[^ǂݍݏItO

	bool VolumeFlag;			//ftHg{[֐tO
	volatile unsigned long LeftVolume;	//obt@Pʂ̎蓮{[
	volatile unsigned long RightVolume;	//obt@Pʂ̎蓮{[

	I_PlayEndCallback *PlayEndCallback;

private:
	bool SetupDevice(int device,const Format &iFormat,
					 long bufsize,int bufnum);

	static void CALLBACK waveOutProc(HWAVEOUT hwo,UINT uMsg,DWORD dwInstance,
									 DWORD dwParam1,DWORD dwParam2);
	static unsigned __stdcall PlayThread(void *woPlayer);

private://
	base_waveOutStreamPlayer(base_waveOutStreamPlayer &);
	void operator=(const base_waveOutStreamPlayer &);
};

//
class waveOutStreamPlayer : public base_waveOutStreamPlayer
{
public:
	waveOutStreamPlayer(void) : base_waveOutStreamPlayer() {}
	virtual ~waveOutStreamPlayer(void) {Terminalize();}

	bool Initialize(int iDevice,const Format &iFormat,
							long iBufsize = 0,int iBufnum = 3);
	bool Initialize(int iDevice,I_InputStream &ioStream,
					long iBufsize = 0,int iBufnum = 3);

	void Terminalize(void);

	virtual int LoadStream(void);
	virtual bool SetBufferVolume(unsigned long iLeft,unsigned long iRight);
private:
	void ChangeBufferVolume(int bufIndex);
};


//{[
class waveOutExVolumeStreamPlayer : public base_waveOutStreamPlayer
{
public:
	waveOutExVolumeStreamPlayer(void)
		: base_waveOutStreamPlayer(),OriginalDataBuffer(NULL)
	{VolumeFlag = true;}

	virtual ~waveOutExVolumeStreamPlayer(void) {Terminalize();}

	bool Initialize(int iDevice,const Format &iFormat,
							long iBufsize = 0,int iBufnum = 3);
	bool Initialize(int iDevice,I_InputStream &ioStream,
					long iBufsize = 0,int iBufnum = 3);
	void Terminalize(void);

	virtual int LoadStream(void);
	virtual bool SetBufferVolume(unsigned long iLeft,unsigned long iRight);
private:
	virtual void ChangeBufferVolume(int bufIndex);

	void **OriginalDataBuffer;	//{[삳ĂȂf[^obt@
};


class base_waveOutPoolPlayer : public I_waveOutPoolPlayer
{
protected:
	friend class waveOutPlayDriver;
	base_waveOutPoolPlayer(void);
	virtual ~base_waveOutPoolPlayer() {}
public:

	bool IsInitialized(void) const {return hWaveOut != 0;}

	virtual bool Start(void);
	virtual bool Stop(void);
	virtual bool isPlaying(void) const {return PlayFlag;}

	virtual bool SetVolume(unsigned long iLeft,unsigned long iRight)
	{
		if (VolumeFlag)
			return SetBufferVolume(iLeft,iRight);
		else
			return SetDeviceVolume(iLeft,iRight);
	}
	virtual bool GetVolume(unsigned long &oLeft,unsigned long &oRight) const
	{
		if (VolumeFlag)
			return GetBufferVolume(oLeft,oRight);
		else
			return GetDeviceVolume(oLeft,oRight);
	}

	virtual Format GetFormat(void) const {return PlayFormat;}

	virtual void SetLoopFlag(bool loop);
	virtual bool GetLoopFlag(void) const {return LoopFlag;}

	virtual long GetPlayPosition(void);

	virtual unsigned long GetBufferSize(void) const
	{
		return WaveHdr.dwBufferLength;
	}

	virtual bool SetDeviceVolume(unsigned long iLeft,unsigned long iRight);
	virtual bool GetDeviceVolume(unsigned long &oLeft,unsigned long &oRight) const;

	virtual bool &BufferVolumeFlag(void) {return VolumeFlag;}

protected:
	bool base_Initialize(int device,const Format &iFormat,long bufsize);
	bool base_Initialize(I_PoolPlayer &iDuplicateSrc);

	void base_Terminalize(void);

	struct SharedBuffer	//Duplicate邽߂̋Lobt@NX
	{
		void *Buffer;
		int ShareCount;
	};

private:
	bool OpenDevice(unsigned int deviceID,const Format &format);

protected:
	HWAVEOUT hWaveOut;		//ĐfoCX
	SharedBuffer *Buffer;	//obt@

	WAVEHDR WaveHdr;	//
	WAVEHDR LoopHdr;	//[vp

	Format PlayFormat;

	bool PlayFlag,LoopFlag;
	bool VolumeFlag;			//ftHg{[֐tO

private://
	base_waveOutPoolPlayer(base_waveOutPoolPlayer &);
	void operator=(const base_waveOutPoolPlayer &);
};


//
class waveOutPoolPlayer : public base_waveOutPoolPlayer
{
public:
	waveOutPoolPlayer(void) : base_waveOutPoolPlayer() {}
	virtual ~waveOutPoolPlayer(void) {Terminalize();}

	virtual const I_PlayDriver *GetDriver(void);

	bool Initialize(int iDevice,const Format &iFormat,long iBufSize);
	bool Initialize(int iDevice,I_SeekableInputStream &iStream);
	bool Initialize(I_PoolPlayer &iDuplicateSrc);

	void Terminalize(void) {base_Terminalize();}

	virtual bool LoadStream(I_InputStream &ioStream);

	virtual bool SetBufferVolume(unsigned long iLeft,unsigned long iRight)
	{return false;}
	virtual bool GetBufferVolume(unsigned long &oLeft,unsigned long &oRight) const
	{return false;}
};



//{[
class waveOutExVolumePoolPlayer : public base_waveOutPoolPlayer
{
public:
	waveOutExVolumePoolPlayer(void)
	  : base_waveOutPoolPlayer(),
		LeftVolume(0xFFFF),RightVolume(0xFFFF)
	{BufferVolumeFlag() = true;}
	~waveOutExVolumePoolPlayer() {Terminalize();}

	virtual const I_PlayDriver *GetDriver(void);

	bool Initialize(int iDevice,const Format &iFormat,long iBufSize);
	bool Initialize(int iDevice,I_SeekableInputStream &iStream);
	bool Initialize(I_PoolPlayer &iDuplicateSrc);
	void Terminalize(void);

	virtual bool LoadStream(I_InputStream &ioStream);

	virtual bool SetBufferVolume(unsigned long iLeft,unsigned long iRight);
	virtual bool GetBufferVolume(unsigned long &oLeft,unsigned long &oRight) const
	{
		oRight = RightVolume;oLeft = LeftVolume;
		return IsInitialized();
	}

private:
	volatile unsigned long LeftVolume;	//obt@Pʂ̎蓮{[
	volatile unsigned long RightVolume;	//obt@Pʂ̎蓮{[

	void ChangeBufferVolume(void);
};

}//namespace Sound
}//namespace Juna

#endif
